#define BUFSIZE 8192
#define STRLEN 8192

#define TCP_SEND_SOCKBUFSIZE 8192  /* socket buffer size */
#define TCP_RECEIVE_SOCKBUFSIZE 8192  /* socket buffer size */
#define TCP_NAGLE 0

struct connection_state *tcp_cs=NULL ;

void tcp_signal_handler(int flag)
{
  fprintf(stderr,"ttp_tcp: interface %s is exiting after receiving SIGQUIT\n",tcp_cs->name) ;
  exit(-1) ;
}

/* returns socket fd */
int init_tcp_client_socket(char *hostname, int port)
{
  int result= -1 ;
  int pb_socket, opt, handshake ;
  struct sockaddr_in locaddr, remaddr ;
  char temps[STRLEN] ;
  char message[STRLEN] ;
  char hostip[STRLEN] ;
  struct hostent *hostaddr ;

  if((pb_socket = socket(AF_INET,SOCK_STREAM,0))<0)
    {
      perror("init_tcp_client_socket: socket") ;
      if(errno==EMFILE)
	{
	  fprintf(stderr,"Too many open fds? What's wrong?...\n") ;
	  exit(-1) ;
	}
      return(-4) ;
    }

  locaddr.sin_family = AF_INET ;
  locaddr.sin_addr.s_addr = htonl(INADDR_ANY) ;
  locaddr.sin_port = htons(port) ;

  opt = 1 ;
  if(setsockopt(pb_socket, SOL_SOCKET, SO_REUSEADDR,(char *) &opt, sizeof(opt))==-1)
    {
      perror("setsockopt (1)") ;
      close(pb_socket) ;
      return(-4) ;
    }

  opt = TCP_SEND_SOCKBUFSIZE ;
  if(setsockopt(pb_socket, SOL_SOCKET, SO_SNDBUF,(char *) &opt, sizeof(opt))==-1)
    {
      perror("setsockopt (2)") ;
      close(pb_socket) ;
      return(-4) ;
    }

  opt = TCP_RECEIVE_SOCKBUFSIZE ;
  if(setsockopt(pb_socket, SOL_SOCKET, SO_RCVBUF,(char *) &opt, sizeof(opt))==-1)
    {
      perror("setsockopt (3)") ;
      close(pb_socket) ;
      return(-4) ;
    }

  if(!TCP_NAGLE)
    {
      opt = 1 ;  /* disable nagle algorithm */
      if(setsockopt(pb_socket, IPPROTO_TCP, TCP_NODELAY,(char *) &opt, sizeof(opt))==-1)
	{
	  perror("setsockopt (4)") ;
	  close(pb_socket) ;
	  return(-4) ;
	}
    }
      
  opt = STRLEN ;
  if(getsockopt(pb_socket,SOL_SOCKET,SO_SNDBUF,message,&opt)==-1)
    {
      perror("init_tcp_client_socket: getsockopt (1)") ;
      close(pb_socket) ;
      return(-4) ;
    }

  opt=*((int *) message) ;

  opt = STRLEN ;
  if(getsockopt(pb_socket,SOL_SOCKET,SO_RCVBUF,message,&opt)==-1)
    {
      perror("init_tcp_client_socket: getsockopt (2)") ;
      close(pb_socket) ;
      return(-4) ;
    }
  opt=*((int *) message) ;

  hostip[0]=0 ;

  if((remaddr.sin_addr.s_addr = inet_addr(hostname)) != -1)  /* user gave us an IP address */
    {
      sprintf(hostip,"%s",hostname) ;
    }
  else  /* resolve hostname */
    {
      hostaddr=gethostbyname(hostname) ;
      if(hostaddr!=NULL)
        {
          sprintf(hostip,"%d.%d.%d.%d",
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[0]),
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[1]),
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[2]),
                  (((unsigned char *) (hostaddr->h_addr_list[0]))[3])) ;
        }
      else
        {
          int ipb0, ipb1, ipb2, ipb3 ;
          if(4==sscanf(hostname,"%d.%d.%d.%d",&ipb0,&ipb1,&ipb2,&ipb3))
            {
	      printf("init_tcp_client_socket: cannot resolve host name, trying literal IP address\n") ;
              sprintf(hostip,"%d.%d.%d.%d",ipb0,ipb1,ipb2,ipb3) ;
            }
        }
      if(hostip[0]==0)
        {
	  fprintf(stderr,"init_tcp_client_socket: cannot resolve host name : %s\n",hostname) ;
	  close(pb_socket) ;
          return(-2) ;      
        }

      /* if(inet_aton(hostip,&(remaddr.sin_addr))!=1) */
      if((remaddr.sin_addr.s_addr = inet_addr(hostip)) == -1)
        {
	  fprintf(stderr,"init_tcp_client_socket: invalid host name\n") ;
	  close(pb_socket) ;
          return(-2) ;
        }
    }

  remaddr.sin_family = AF_INET ;
  remaddr.sin_port = htons(port) ;

  sprintf(temps,"init_tcp_client_socket: attempting to connect to %s:%d",hostname,port) ;

  if(connect(pb_socket,(struct sockaddr*) &remaddr,sizeof(struct sockaddr_in))!=0)
    {
      sprintf(temps,"init_tcp_client_socket: connect to %s:%d failed",hostname,port) ;
      perror(temps) ;
      close(pb_socket) ;
    }
  else
    {
      sprintf(temps,"init_tcp_client_socket: established connection to %s:%d",hostname,port) ;
      result=pb_socket ;
    }
  return(result) ;
}

/* returns socket fd */
int init_tcp_server_socket(int local_port)
{
  int tcp_socket ;
  int opt ;
  char message[512] ;
  struct sockaddr_in locaddr ;

  fprintf(stderr,"ttp_tcp (server): initializing\n") ;

  if((tcp_socket = socket(AF_INET,SOCK_STREAM,0))<0)
    {
      perror("ttp_tcp: socket") ;
      exit(-1) ;
    }

  locaddr.sin_family = AF_INET ;
  locaddr.sin_addr.s_addr = htonl(INADDR_ANY) ;
  locaddr.sin_port = htons(local_port) ;

  opt = 1 ;  /* enable */
  if(setsockopt(tcp_socket, SOL_SOCKET, SO_REUSEADDR,(char *) &opt, sizeof(opt))==-1)
    {
      perror("ttp_tcp: setsockopt (0)") ;
      exit(-1) ;
    }

  opt = 0 ;  /* disable */
  /* keepalive causes the connection to die after approx 132 minutes of no response */
  if(setsockopt(tcp_socket, SOL_SOCKET, SO_KEEPALIVE,(char *) &opt, sizeof(opt))==-1)
    {
      perror("ttp_tcp: setsockopt (1)") ;
      exit(-1) ;
    }

  opt = TCP_SEND_SOCKBUFSIZE ;
  if(setsockopt(tcp_socket, SOL_SOCKET, SO_SNDBUF,(char *) &opt, sizeof(opt))==-1)
    {
      perror("ttp_tcp: setsockopt (2)") ;
      exit(-1) ;
    }

  opt = TCP_RECEIVE_SOCKBUFSIZE ;
  if(setsockopt(tcp_socket, SOL_SOCKET, SO_RCVBUF,(char *) &opt, sizeof(opt))==-1)
    {
      perror("ttp_tcp: setsockopt (3)") ;
      exit(-1) ;
    }
  
  if(!TCP_NAGLE)
    {
      opt = 1 ;  /* disable nagle algorithm */
      if(setsockopt(tcp_socket, IPPROTO_TCP, TCP_NODELAY,(char *) &opt, sizeof(opt))==-1)
        {
          perror("ttp_tcp: setsockopt (4)") ;
          exit(-1) ;
        }
    }

  opt = 128 ;
  if(getsockopt(tcp_socket,SOL_SOCKET,SO_SNDBUF,message,&opt)==-1)
    {
      perror("ttp_tcp: getsockopt (1)") ;
      exit(-1) ;
    }
  opt=*((int *) message) ;
  fprintf(stderr,"ttp_tcp: Send buffer is %d bytes.\n",opt) ;

  opt = 128 ;
  if(getsockopt(tcp_socket,SOL_SOCKET,SO_RCVBUF,message,&opt)==-1)
    {
      perror("ttp_tcp: getsockopt (2)") ;
      exit(-1) ;
    }
  opt=*((int *) message) ;
  fprintf(stderr,"ttp_tcp: Receive buffer is %d bytes.\n",opt) ;

  if(bind(tcp_socket,(struct sockaddr *) &locaddr,sizeof(locaddr))<0)
    {
      perror("ttp_tcp: bind") ;
    }

  if(listen(tcp_socket,MAX_TCP_CONNECTIONS)<0)
    perror("ttp_tcp: listen") ;

  fprintf(stderr,"ttp_tcp: listening on local port %d\n",local_port) ;

  return(tcp_socket) ;
}

void run_tcp_server(struct connection_state **transients, int max_transients, int local_port)
{
  struct connection_state *cs ;
  int done=0 ;
  int i, r, j, server_fd, maxfd, buflen, namelen, audience ;
  fd_set readfds ;
  char buf[8192] ;
  struct timeval timeout ;
  struct sockaddr_in remaddr ;
  int num_transients=0, which_transient ;
  struct timeval now ;
  unsigned long rs ;
  int *tfd ;
  long *input_seed, *output_seed ;

  gettimeofday(&now,NULL) ;
  rs=((unsigned long) getpid())*now.tv_usec + now.tv_sec ;
  rs^=(unsigned long) &rs ;
  srand48(rs) ;

  tfd=(int *)calloc(max_transients,sizeof(int)) ;
  input_seed=(long *)calloc(max_transients,sizeof(long)) ;
  output_seed=(long *)calloc(max_transients,sizeof(long)) ;

  server_fd=init_tcp_server_socket(local_port) ;

  fprintf(stderr,"ttp_tcp (server): done initializing...\n") ;

  for(;!done;)
    {
      /* fprintf(stderr,"ttp_tcp (server: %s): get ready to select...\n",transients[0]->name) ; */
      maxfd= -1 ;
      FD_ZERO(&readfds) ;
      for(i=0;i<max_transients;i++)
	{
	  cs=transients[i] ;
	  if(cs->active)
	    {
	      /* fprintf(stderr,"ttp_tcp: transient %d is active, checking for input on fd %d\n",i,tfd[i]) ; */

	      FD_SET(tfd[i],&readfds) ;
	      if(tfd[i]>maxfd)
		maxfd=tfd[i] ;
	      
	      FD_SET(cs->ttp_to_interface[0],&readfds) ;
	      if(cs->ttp_to_interface[0]>maxfd)
		maxfd=cs->ttp_to_interface[0] ;

	      FD_SET(cs->ttp_death_signal[0],&readfds) ;
	      if(cs->ttp_death_signal[0]>maxfd)
		maxfd=cs->ttp_death_signal[0] ;
	    }
	}
      FD_SET(server_fd,&readfds) ;
      if(server_fd>maxfd)
	maxfd=server_fd ;

      timeout.tv_sec=64 ;
      timeout.tv_usec=0 ;

      /* fprintf(stderr,"ttp_tcp (server): selecting... (maxfd=%d)\n",maxfd) ; */
      r=select(maxfd+1,&readfds,NULL,NULL,&timeout) ;

      if(r<0)
	perror("ttp_tcp: select") ;
      else
	{
	  for(i=0;i<max_transients;i++)
	    {
	      cs=transients[i] ;
	      if(cs->active)
		{
		  if(FD_ISSET(tfd[i],&readfds))
		    {
		      /* fprintf(stderr,"ttp_tcp_server (%s): data available from transient %d\n",cs->name,i) ; */
		      if((buflen=syswrap_read(tfd[i],buf,8192))>0)
			{
			  /* fprintf(stderr,"%s transient %d has input seed 0x%X and output seed 0x%X\n",cs->name,i,input_seed[i],output_seed[i]) ;  */
			  /* decrypt buffer */
			  for(j=0;j<buflen;j++)
			    {
			      /* fprintf(stderr,"change byte %x (%c)",buf[j],buf[j]) ; */
			      buf[j]^=(0xFF & trux_simple_rand(&(input_seed[i]))) ;
			      /* fprintf(stderr," to %x (%c) from tcp transient %d\n",buf[j],buf[j],i) ; */
			    }
			  
			  /* fprintf(stderr,"ttp_tcp_server: relay %d bytes from %s to ttp\n",buflen,cs->name) ; */
			  sure_write(NULL,cs->interface_to_ttp[1],buf,buflen,0) ;
			}
		      else
			{
			  fprintf(stderr,"ttp_tcp_server (%s): transient %d selected but no data ready\n",cs->name,i) ;
			  /* fprintf(stderr,"tcp interface %s becomes inactive\n",cs->name) ; */
			  cs->active=0 ;
			  close(tfd[i]) ;
			  *((int *)buf)=htonl(TTP_INACTIVE) ;
			  sure_write(NULL,cs->interface_to_ttp[1],buf,sizeof(int),0) ;
			  num_transients-- ;
			}
		    }
		  if(FD_ISSET(cs->ttp_death_signal[0],&readfds))
		    {
		      if((buflen=syswrap_read(cs->ttp_death_signal[0],buf,8192))>0)
			{
			  fprintf(stderr,"ttp_tcp_server (%s): death signal for %d\n",cs->name,i) ;
			  cs->active=0 ;
			  close(tfd[i]) ;
			  *((int *)buf)=htonl(TTP_INACTIVE) ;
			  sure_write(NULL,cs->interface_to_ttp[1],buf,sizeof(int),0) ;
			  num_transients-- ;
			}
		    }
		  if(FD_ISSET(cs->ttp_to_interface[0],&readfds))
		    {
		      /* fprintf(stderr,"ttp_tcp_server (%s): data available from ttp for transient %d\n",cs->name,i) ; */
		      if((buflen=syswrap_read(cs->ttp_to_interface[0],buf,8192))>0)
			{
			  /* fprintf(stderr,"ttp_tcp_server: read %d bytes from ttp, relay to transient %d\n",buflen,i) ; */
			  sure_write(&(output_seed[i]),tfd[i],buf,buflen,0) ;  /* encrypt */
			}
		    }
		}
	    }
	  
	  if(FD_ISSET(server_fd,&readfds))
	    {
	      fprintf(stderr,"ttp_tcp: new connection on %s ...\n",transients[0]->name) ;
	      cs=NULL ;
	      for(i=0;(cs==NULL)&&(i<max_transients);)
		if(!(transients[i]->active))
		  cs=transients[i] ;
		else
		  i++ ;
	      if(cs!=NULL)
		{
		  namelen=sizeof(struct sockaddr_in) ;
		  fprintf(stderr,"accepting...\n") ;
		  for(;(audience=accept(server_fd,(struct sockaddr *) &remaddr,&namelen))<0;)
		    perror("ttp_tcp: accept") ;
		  fprintf(stderr,"%s acceped connection from %s (%s)\n",timestamps(),inet_ntoa(remaddr.sin_addr),cs->name) ;

		  if(diffie_hellman_handshake(audience, &(input_seed[i]), &(output_seed[i]), 1))
		    {
		      close(audience) ;
		    }
		  else
		    {
		      fprintf(stderr,"tcp interface %s transient %d becomes active with fd %d\n",cs->name,i,audience) ;
		      /* fprintf(stderr,"tcp interface %s transient %d has input seed 0x%lX and output seed 0x%lX\n",cs->name,i,input_seed[i],output_seed[i]) ; */
		      cs->active=1 ;
		      tfd[i]=audience ;
		      *((int *)buf)=htonl(TTP_ACTIVE) ;
		      sure_write(NULL,cs->interface_to_ttp[1],buf,sizeof(int),0) ;
		      num_transients++ ;
		    }
		}
	    }
	}
    }
  fprintf(stderr,"ttp_tcp server %s done\n",transients[0]->name) ;
}

void run_tcp_client(struct connection_state *cs,char *remote_hostname, int remote_port, int local_port)
{
  int done ;
  int r, j, fd, maxfd, buflen, namelen ;
  long input_seed, output_seed ;
  fd_set readfds ;
  char buf[8192] ;
  struct timeval timeout ;
  struct sockaddr_in remaddr ;

  for(;;)
    {
      fprintf(stderr,"trying to connect to %s\n",cs->name) ;
      fd=init_tcp_client_socket(remote_hostname,remote_port) ;
      if(fd>=0)
	{
	  if(diffie_hellman_handshake(fd,&input_seed,&output_seed,0))
	    {
	      fprintf(stderr,"diffie hellman handshake failed\n") ;
	      close(fd) ;
	      fd= -1 ;
	    }
	}
      if(fd>=0)
	{
	  fprintf(stderr,"tcp interface %s becomes active\n",cs->name) ;
	  /* fprintf(stderr,"tcp interface %s has input seed 0x%lX and output seed 0x%lX\n",cs->name,input_seed,output_seed) ; */
	  cs->active=1 ;
	  *((int *)buf)=htonl(TTP_ACTIVE) ;
	  sure_write(NULL,cs->interface_to_ttp[1],buf,sizeof(int),0) ;
	  
	  for(done=0;!done;)
	    {
	      FD_ZERO(&readfds) ;
	      maxfd=fd ;
	      FD_SET(fd,&readfds) ;
	      FD_SET(cs->ttp_to_interface[0],&readfds) ;
	      if(maxfd<cs->ttp_to_interface[0])
		maxfd=cs->ttp_to_interface[0] ;
	      timeout.tv_sec=64 ;
	      timeout.tv_usec=0 ;
	      
	      /* fprintf(stderr,"ttp_tcp (client): selecting...\n") ; */
	      r=select(maxfd+1,&readfds,NULL,NULL,&timeout) ;

	      if(FD_ISSET(cs->ttp_to_interface[0],&readfds))
		if((buflen=syswrap_read(cs->ttp_to_interface[0],buf,8192))>0)
		  {
		    sure_write(&output_seed,fd,buf,buflen,0) ;  /* encrypt */
		    /* fprintf(stderr,"tcp client reads %d bytes from ttp and relays to tcp\n",buflen) ; */
		  }

	      if(FD_ISSET(fd,&readfds))
                {
		  buflen=syswrap_read(fd,buf,8192) ;
		  if(buflen>0)
		    {
		      /* decrypt buffer */
		      for(j=0;j<buflen;j++)
			{
			  /* fprintf(stderr,"change byte %x (%c)",buf[j],buf[j]) ; */
			  buf[j]^=(0xFF & trux_simple_rand(&input_seed)) ;
			  /* fprintf(stderr," to %x (%c) from tcp client %s\n",buf[j],buf[j],cs->name) ; */
			}

		      /* relay buffer */
		      sure_write(NULL,cs->interface_to_ttp[1],buf,buflen,0) ;
		      /* fprintf(stderr,"tcp client reads %d bytes and relays to ttp\n",buflen) ; */
		    }
		  else
		    {
		      fprintf(stderr,"ttp_tcp: %s: selected, but %d bytes read\n",cs->name,buflen) ;
		      perror("read") ;
		      done=1 ;
		    }
                }
	    }
	  close(fd) ;
	  /* fprintf(stderr,"tcp interface %s becomes inactive\n",cs->name) ; */
	  cs->active=0 ;
	  *((int *)buf)=htonl(TTP_INACTIVE) ;
	  sure_write(NULL,cs->interface_to_ttp[1],buf,sizeof(int),0) ;
	}
      sleep(16) ;
      fprintf(stderr,"ttp_tcp: reinitializing interface: %s (mypid=%d)\n",cs->name,getpid()) ;
    }
}

void init_tcp_server(struct connection_state **transients, int max_transients, int local_port)
{
  int i, pid ;

  for(i=0;i<max_transients;i++)
    {
      if(pipe(transients[i]->ttp_to_interface) ||
	 pipe(transients[i]->interface_to_ttp) ||
	 pipe(transients[i]->ttp_death_signal))
	{
	  perror("pipe creation failed") ;
	  exit(-1) ;
	}
    }

  pid=fork() ;
  for(i=0;i<max_transients;i++)
    transients[i]->pid=pid ;
  if(!pid)
    {
      signal(SIGQUIT,tcp_signal_handler) ;
      run_tcp_server(transients, max_transients, local_port) ;
    }
}

void init_tcp_client(struct connection_state *cs, char *remote_hostname, int remote_port, int local_port)
{
  int i ;

  pipe(cs->ttp_to_interface) ;
  pipe(cs->interface_to_ttp) ;

  cs->pid=fork() ;
  if(!cs->pid)
    {
      tcp_cs=cs ;
      signal(SIGQUIT,tcp_signal_handler) ;
      run_tcp_client(cs,remote_hostname, remote_port, local_port) ;
    }
}

